/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or (at
 *  your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 *  MA 02110-1301, USA.
 *
 *  Copyright © 2006 Pierre Habouzit
 */

/** \addtogroup mutt_strings */
/*@{*/

/** \file str.c
 * \brief Madmutt string API module implementation.
 * \author Pierre Habouzit <madcoder@debian.org>
 */

#include "lib-lib.h"

#ifndef __doxygen_skip__
#define XX 255
unsigned char const __m_strdigits[128] = {
    XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX,
    XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX,
    XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX,
     0,  1,  2,  3,  4,  5,  6,  7,  8,  9, XX, XX, XX, XX, XX, XX,
    XX, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24,
    25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, XX, XX, XX, XX, XX,
    XX, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24,
    25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, XX, XX, XX, XX, XX,
};
#undef XX

#define XX -1
signed char const __m_b64digits[128] = {
    XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX,
    XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX,
    XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, XX, 62, XX, XX, XX, 63,
    52, 53, 54, 55, 56, 57, 58, 59, 60, 61, XX, XX, XX, XX, XX, XX,
    XX,  0,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10, 11, 12, 13, 14,
    15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, XX, XX, XX, XX, XX,
    XX, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40,
    41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 51, XX, XX, XX, XX, XX
};
#undef XX

char const __m_b64chars[64] = {
    'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L', 'M', 'N', 'O',
    'P', 'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X', 'Y', 'Z', 'a', 'b', 'c', 'd',
    'e', 'f', 'g', 'h', 'i', 'j', 'k', 'l', 'm', 'n', 'o', 'p', 'q', 'r', 's',
    't', 'u', 'v', 'w', 'x', 'y', 'z', '0', '1', '2', '3', '4', '5', '6', '7',
    '8', '9', '+', '/'
};

char const __m_b36chars_lower[36] = {
    '0', '1', '2', '3', '4', '5', '6', '7', '8', '9',
    'a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j',
    'k', 'l', 'm', 'n', 'o', 'p', 'q', 'r', 's', 't',
    'u', 'v', 'w', 'x', 'y', 'z'
};

char const __m_b36chars_upper[36] = {
    '0', '1', '2', '3', '4', '5', '6', '7', '8', '9',
    'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J',
    'K', 'L', 'M', 'N', 'O', 'P', 'Q', 'R', 'S', 'T',
    'U', 'V', 'W', 'X', 'Y', 'Z'
};
#endif

/** \brief safe strcpy.
 *
 * Copies at most <tt>n-1</tt> characters from \c src into \c dst, always
 * adding a final \c \\0 in \c dst.
 *
 * \param[in]  dst      destination buffer.
 * \param[in]  n        size of the buffer. Negative sizes are allowed.
 * \param[in]  src      source string.
 *
 * \return \c src \e length. If this value is \>= \c n then the copy was
 *         truncated.
 */
ssize_t m_strcpy(char *dst, ssize_t n, const char *src)
{
    ssize_t len = m_strlen(src);

    if (n > 0) {
        ssize_t dlen = MIN(n - 1, len);
        memcpy(dst, src, dlen);
        dst[dlen] = '\0';
    }

    return len;
}

/** \brief safe limited strcpy.
 *
 * Copies at most min(<tt>n-1</tt>, \c l) characters from \c src into \c dst,
 * always adding a final \c \\0 in \c dst.
 *
 * \param[in]  dst      destination buffer.
 * \param[in]  n        size of the buffer. Negative sizes are allowed.
 * \param[in]  src      source string.
 * \param[in]  l        maximum number of chars to copy.
 *
 * \return minimum of  \c src \e length and \c l.
 */
ssize_t m_strncpy(char *dst, ssize_t n, const char *src, ssize_t l)
{
    ssize_t len = MIN(m_strlen(src), l);

    if (n > 0) {
        ssize_t dlen = MIN(n - 1, len);
        memcpy(dst, src, dlen);
        dst[dlen] = '\0';
    }

    return len;
}

char *m_strrtrim(char *s)
{
    ssize_t len = m_strlen(s);

    while (len > 1 && ISSPACE(s[len - 1]))
        s[--len] = '\0';

    return s + len;
}

const char *m_stristrn(const char *haystack, const char *needle, ssize_t nlen)
{
    int nc;

    if (!nlen)
        return haystack;

    nc = tolower(*needle);
    for (;;) {
        int c = tolower(*haystack);

        if (c != nc) {
            if (c == '\0')
                return NULL;
        } else {
            ssize_t i;

            /* compare the rest of needle */
            for (i = 1;; i++) {
                if (i == nlen)
                    return haystack;
                if (c == '\0')
                    return NULL;
                c = tolower(haystack[i]);
                if (c != tolower(needle[i]))
                    break;
            }
        }

        haystack++;
    }
}

/** \brief \c NULL resistant strcasecmp.
 * \param[in]  a     the first string.
 * \param[in]  b     the second string.
 * \return <tt>strcasecmp(a, b)</tt>, and treats \c NULL strings like \c ""
 *         ones, as if we were in the C locale.
 */
int ascii_strcasecmp(const char *a, const char *b)
{
    if (a == b)
        return 0;
    if (!a)
        return -1;
    if (!b)
        return 1;

    while (*a || *b) {
        int i;
        if ((i = ascii_tolower(*a++) - ascii_tolower(*b++)))
            return i;
    }

    return 0;
}

/** \brief \c NULL resistant strncasecmp.
 * \param[in]  a     the first string.
 * \param[in]  b     the second string.
 * \param[in]  n     the number of maximum chars to compare.
 * \return <tt>strncasecmp(a, b)</tt>, and treats \c NULL strings like \c ""
 *         ones, as if we were in the C locale.
 */
int ascii_strncasecmp(const char *a, const char *b, ssize_t n)
{
    if (a == b)
        return 0;
    if (!a)
        return -1;
    if (!b)
        return 1;

    while ((*a || *b) && n > 0) {
        int i;
        if ((i = ascii_tolower(*a++) - ascii_tolower(*b++)))
            return i;
        n--;
    }

    return 0;
}

/** \brief Retutnrs the on-screen width of the string, multi-byte aware.
 *
 * The function assumes that at rendering time, any non printable character or
 * any unrecognized multibyte sequence will be replaced with a 1-cell wide
 * char.
 *
 * \param[in]  s   the string;
 * \return the width of \c s.
 */
ssize_t m_strwidth (const char *s)
{
    const ssize_t len = m_strlen(s);
    mbstate_t mbstate;
    ssize_t w = 0, pos = 0;

    p_clear(&mbstate, 1);

    while (pos < len) {
        wchar_t wc;
        ssize_t nb;

        nb = mbrtowc(&wc, s + pos, len - pos, &mbstate);
        switch (nb) {
          case -2:
            return w + 1;

          case -1: /* assume we will replace that char with a ? later */
            wc = '?';
            pos ++;
            break;

          default:
            w   += iswprint(wc) ? wcwidth(wc) : 1;
            pos += nb;
        }
    }

    return w;
}

ssize_t m_strformat(char *dst, ssize_t dlen, int width, const char *fmt,
                    format_t *callback, anytype cdata, format_flag flags)
{
    ssize_t pos = flags & M_FORMAT_ARROWCURSOR ? 3 : 0;

    m_strpad(dst, dlen, '\0', pos + 1);
    if (!fmt)
        return pos;

    while (*fmt) {
        int ch;

        if (*fmt == '%') {
            char ifstr[STRING], elstr[STRING], prefix[STRING];

            *ifstr = *elstr = *prefix = '\0';

            if (*++fmt == '%') {
                pos += m_strputc(dst + pos, dlen - pos, *fmt++);
                continue;
            }

            if (*fmt == '?') {
                flags |= M_FORMAT_OPTIONAL;
                fmt++;
            } else {
                ssize_t pfxlen;
                flags &= ~M_FORMAT_OPTIONAL;

                /* eat the format string */
                pfxlen = strspn(fmt, "0123456789.-");
                m_strncpy(prefix, sizeof(prefix), fmt, pfxlen);
                fmt   += pfxlen;
            }

            /* save the character to switch on */
            if (!(ch = *fmt++))
                break;

            if (flags & M_FORMAT_OPTIONAL) {
                ssize_t iflen;
                const char *p;

                if (*fmt++ != '?')
                    break;                /* bad format */

                /* eat the `if' part of the string */
                iflen = strcspn(fmt, "?&");
                m_strncpy(ifstr, ssizeof(ifstr), fmt, iflen);
                fmt  += iflen;

                /* eat the `else' part of the string (optional) */
                if (*fmt == '&')
                    fmt++;                /* skip the & */

                p = m_strchrnul(fmt, '?');
                m_strncpy(elstr, ssizeof(elstr), fmt, p - fmt);
                fmt = p;

                if (!*fmt++)              /* move past the trailing `?' */
                    break;                /* bad format */
            }

            switch (ch) {
                char lower, nodots, buf[LONG_STRING];

              case '>':                 /* right justify to EOL */
                width -= m_strwidth(dst);

                ch = *fmt++;            /* pad char */

                if (width > 0) {
                    m_strformat(buf, sizeof(buf), 0, fmt, callback, cdata, flags);
                    width -= m_strwidth(buf);
                    pos += m_strpad(dst + pos, dlen - pos, ch, width);
                    pos += m_strcpy(dst + pos, dlen - pos, buf);
                }
                return pos;             /* skip rest of input */

              case '|':                 /* pad to EOL */
                width -= m_strwidth(dst);
                return pos + m_strpad(dst + pos, dlen - pos, *fmt, width);

              default:
                lower = nodots = 0;

                while (ch == '_' || ch == ':') {
                    lower  |= ch == '_';
                    nodots |= ch == ':';
                    ch = *fmt++;
                }

                /* use callback function to handle this case */
                fmt = callback(buf, sizeof (buf), ch, fmt, prefix,
                               ifstr, elstr, cdata, flags);

                if (lower)
                    m_strtolower(buf);

                if (nodots) {
                    char *p;

                    for (p = buf; *p; p++) {
                        if (*p == '.')
                            *p = '_';
                    }
                }

                pos += m_strcpy(dst + pos, dlen - pos, buf);
                break;
            }
            continue;
        }

        if (*fmt == '\\') {
            if (!*++fmt)
                break;
            switch ((ch = *fmt++)) {
              case 'n': pos += m_strputc(dst + pos, dlen - pos, '\n'); break;
              case 't': pos += m_strputc(dst + pos, dlen - pos, '\t'); break;
              case 'r': pos += m_strputc(dst + pos, dlen - pos, '\r'); break;
              case 'f': pos += m_strputc(dst + pos, dlen - pos, '\f'); break;
              case 'v': pos += m_strputc(dst + pos, dlen - pos, '\v'); break;
              default:  pos += m_strputc(dst + pos, dlen - pos, ch);   break;
            }
        } else {
            ssize_t len = strcspn(fmt, "%\\");

            pos += m_strncpy(dst + pos, dlen - pos, fmt, len);
            fmt += len;
        }
    }

    return pos;
}

/*@}*/
